from o3seespy.base_model import OpenSeesObject, OpenSeesMultiCallObject
from o3seespy.opensees_instance import OpenSeesInstance
# from o3seespy.cc import
from o3seespy.exceptions import ModelError

def set_node_mass(osi, node, x_mass, y_mass, rot_mass):
    op_type = 'mass'
    parameters = [node.tag, x_mass, y_mass, rot_mass]
    osi.to_process(op_type, parameters)


class Mass(OpenSeesObject):
    op_base_type = "mass"
    op_type = None

    def __init__(self, osi, node, x_mass, y_mass, rot_mass=None):
        if osi.ndf > 2 and rot_mass is None:
            rot_mass = 0.0
        self.node = node
        self.x_mass = x_mass
        self.y_mass = y_mass
        self.rot_mass = rot_mass
        self._parameters = [self.node.tag, self.x_mass, self.y_mass]
        if self.rot_mass is not None:
            self._parameters.append(self.rot_mass)
        self.to_process(osi)
        
        
class Mass3D(OpenSeesObject):
    op_base_type = "mass"
    op_type = None

    def __init__(self, osi, node, x, y, z, x_rot, y_rot, z_rot):
        self.node = node
        self.x = x
        self.y = y
        self.z = y
        self.x_rot = x_rot
        self.y_rot = y_rot
        self.z_rot = z_rot
        self._parameters = [self.node.tag, self.x, self.y, self.z, self.x_rot, self.y_rot, self.z_rot]
        self.to_process(osi)


def set_equal_dof(osi, node_1, node_2, dof):
    op_type = 'equalDOF'
    parameters = [node_1.tag, node_2.tag,  dof]
    osi.to_process(op_type, parameters)


def set_equal_dofs(osi, node_1, node_2, dofs):
    op_type = 'equalDOF'
    parameters = [node_1.tag, node_2.tag,  *dofs]
    osi.to_process(op_type, parameters)


def set_equal_dofs_mixed(osi, node_1, node_2, num_dof, rcdofs):  # TODO: validate
    op_type = 'equalDOF_Mixed'
    parameters = [node_1.tag, node_2.tag,  num_dof, *rcdofs]
    osi.to_process(op_type, parameters)


class EqualDOF(OpenSeesObject):
    op_base_type = "equalDOF"
    op_type = None

    def __init__(self, osi, r_node, c_node, dofs):
        """
        Construct a constraint where the constrained node `c_node`
         has the same movement as the retained node `r_node`

        :param osi:
        :param r_node: OpenSeesObject.node.Node() or list of
            Retained node
        :param c_node: OpenSeesObject.node.Node() or list of
            Constrained node
        :param dofs:
        """
        self.r_node = r_node
        self.c_node = c_node
        self.dofs = dofs
        self._parameters = [self.r_node.tag, self.c_node.tag, *self.dofs]
        self.to_process(osi)


class EqualDOFMulti(OpenSeesMultiCallObject):
    op_base_type = "equalDOF"
    op_type = None

    def __init__(self, osi, r_node, c_nodes, dofs):
        """
        Construct a constraint where the c_node has the same movement as the r_node

        :param osi:
        :param r_node: OpenSeesObject.node.Node() or list of
            Retained node
        :param c_nodes: list of OpenSeesObject.node.Node()
            Constrained node
        :param dofs:
        """
        self.r_node = r_node
        self.c_nodes = c_nodes
        self.dofs = dofs
        if hasattr(r_node, '__len__'):
            r_nodes = self.r_node
        else:
            r_nodes = [self.r_node for i in range(len(self.c_nodes))]
        self._multi_parameters = []
        for i in range(len(c_nodes)):
            self._multi_parameters.append([r_nodes[i].tag, self.c_nodes[i].tag, *self.dofs])
            self.to_process(osi)


def set_rigid_diaphragm(osi, r_node, cnodes, perp_dir):
    cnode_tags = [x.tag for x in cnodes]
    op_type = 'rigidDiaphragm'
    parameters = [perp_dir, r_node.tag,  *cnode_tags]
    osi.to_process(op_type, parameters)


def set_rigid_link(osi, r_node, c_node, rtype):
    """
    Create a multi-point constraint between nodes.

    Parameters
    ----------
    r_node: OpenSeesObject.node.Node()
        Retained node
    c_node: Node
        Constrained node
    rtype: str
        Either 'bar' or 'beam'

    Returns
    -------

    """
    op_type = 'rigidLink'
    parameters = [rtype, r_node.tag, c_node.tag]
    osi.to_process(op_type, parameters)


class RigidLink(OpenSeesObject):
    op_base_type = 'rigidLink'
    op_type = None
    def __init__(self, osi, rtype, r_node, c_node):
        """
        Create a multi-point constraint between nodes.

        Parameters
        ----------
        rtype: str
            Either 'bar' or 'beam'
        r_node: OpenSeesObject.node.Node()
            Retained node
        c_node: Node
            Constrained node

        Returns
        -------

        """
        self.rtype = rtype
        self.r_node = r_node
        self.c_node = c_node
        self._parameters = [rtype, r_node.tag, c_node.tag]
        self.to_process(osi)


class Fix1DOF(OpenSeesObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, node, x):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        node: OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        """
        self.node = node
        self.x = x
        self._parameters = [self.node.tag, self.x]
        self.to_process(osi)


class Fix1DOFMulti(OpenSeesMultiCallObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, nodes, x):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        nodes: list of OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        """
        self.nodes = nodes
        self.x = x
        self._multi_parameters = []
        for node in self.nodes:
            self._multi_parameters.append([node.tag, self.x])
            self.to_process(osi)


class Fix2DOF(OpenSeesObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, node, x, y):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        node: OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        """
        self.node = node
        self.x = x
        self.y = y
        self._parameters = [self.node.tag, self.x, self.y]
        self.to_process(osi)


class Fix2DOFMulti(OpenSeesMultiCallObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, nodes, x, y, is_none='raise'):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        nodes: list of OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        """
        self.nodes = nodes
        self.x = x
        self.y = y
        self._multi_parameters = []
        for node in self.nodes:
            try:
                self._multi_parameters.append([node.tag, self.x, self.y])
                self.to_process(osi)
            except AttributeError as e:
                if is_none == 'raise':
                    raise e
                else:
                    pass
                

def add_fixity_to_dof(osi, dof, nodes):
    if osi.ndf == 1:
        fn = Fix1DOF
        arr = [1]
    elif osi.ndf == 2:
        fn = Fix2DOF
        arr = [0, 0]
    elif osi.ndf == 3:
        fn = Fix3DOF
        arr = [0, 0, 0]
    elif osi.ndf == 6:
        fn = Fix6DOF
        arr = [0, 0, 0, 0, 0, 0]
    else:
        raise ModelError("'create_fixities_for_dof' only supports ndf=1,2,3,6")
    arr[dof - 1] = 1
    for node in nodes:
        if node is None:
            continue
        try:
            fn(osi, node, *arr)
        except ValueError:
            if osi.state == 3:
                osi.commands = osi.commands[:-1]
            pass


class Fix2DOaFMulti(OpenSeesMultiCallObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, nodes, x, y):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        nodes: list of OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        """
        self.nodes = nodes
        self.x = x
        self.y = y
        self._multi_parameters = []
        for node in self.nodes:
            self._multi_parameters.append([node.tag, self.x, self.y])
            self.to_process(osi)
        

class Fix3DOF(OpenSeesObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, node, x, y, z_rot):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        node: OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        z_rot: int
            Fixity in rotation about z-axis
        """
        self.node = node
        self.x = x
        self.y = y
        self.z_rot = z_rot
        self._parameters = [self.node.tag, self.x, self.y, self.z_rot]
        self.to_process(osi)


class Fix3DOFMulti(OpenSeesMultiCallObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, nodes, x, y, z_rot):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        nodes: list of OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        z_rot: int
            Fixity in rotation about z-axis
        """
        self.nodes = nodes
        self.x = x
        self.y = y
        self.z_rot = z_rot
        self._multi_parameters = []
        for node in self.nodes:
            self._multi_parameters.append([node.tag, self.x, self.y, self.z_rot])
            self.to_process(osi)


class Fix6DOF(OpenSeesObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, node, x, y, z, x_rot, y_rot, z_rot):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        node: OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        z: int
            Fixity in z-direction
        x_rot: int
            Fixity in rotation about x-axis
        y_rot: int
            Fixity in rotation about y-axis
        z_rot: int
            Fixity in rotation about z-axis
        """
        self.node = node
        self.x = x
        self.y = y
        self.z = z
        self.x_rot = x_rot
        self.y_rot = y_rot
        self.z_rot = z_rot
        self._parameters = [self.node.tag, self.x, self.y, self.z, self.x_rot, self.y_rot, self.z_rot]
        self.to_process(osi)


class Fix6DOFMulti(OpenSeesMultiCallObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, nodes, x, y, z, x_rot, y_rot, z_rot):
        """
        Create a homogeneous SP constraint.

        Parameters
        ----------
        osi: OpenSeesInstance
        nodes: list of OpenSeesObject.node.Node()
        x: int
            Fixity in x-direction
        y: int
            Fixity in y-direction
        z: int
            Fixity in z-direction
        x_rot: int
            Fixity in rotation about x-axis
        y_rot: int
            Fixity in rotation about y-axis
        z_rot: int
            Fixity in rotation about z-axis
        """
        self.nodes = nodes
        self.x = x
        self.y = y
        self.z = z
        self.x_rot = x_rot
        self.y_rot = y_rot
        self.z_rot = z_rot
        self._multi_parameters = []
        for node in self.nodes:
            self._multi_parameters.append([node.tag, self.x, self.y, self.z, self.x_rot, self. y_rot, self.z_rot])
            self.to_process(osi)


class Fix(OpenSeesObject):
    op_base_type = "fix"
    op_type = None

    def __init__(self, osi, node, fixities):
        """
        Create a homogeneous SP constraint.

        """
        self.node = node
        self.fixities = fixities
        self._parameters = [self.node.tag, *self.fixities]
        self.to_process(osi)


class Load(OpenSeesObject):
    op_base_type = "load"
    op_type = None

    def __init__(self, osi, node, load_values):
        self.node = node
        self.load_values = load_values

        self._parameters = [self.node.tag, *self.load_values]
        self.to_process(osi)


class EleLoad2DPoint(OpenSeesObject):
    op_base_type = "eleLoad"
    op_type = None

    def __init__(self, osi, ele, p_y, x, p_x=None):
        """
        Type of load is 'beamPoint'

        x: float
            Position of load as a fraction of element length from node i
        """
        self.ele_tag = ele.tag
        self.x = float(x)
        self.p_y = float(p_y)
        self.p_x = p_x

        self._parameters = ['-ele', self.ele_tag, '-type', '-beamPoint', self.p_y, self.x]
        if self.p_x is not None:
            self._parameters.append(float(self.p_x))
        self.to_process(osi)


class EleLoad2DUniform(OpenSeesObject):
    op_base_type = "eleLoad"
    op_type = None

    def __init__(self, osi, ele, w_y, w_x=None):
        """
        Type of load is 'beamUniform'
        """
        self.ele_tag = ele.tag
        self.w_y = float(w_y)
        self.w_x = w_x

        self._parameters = ['-ele', self.ele_tag, '-type', '-beamUniform', self.w_y]
        if self.w_x is not None:
            self._parameters.append(float(self.w_x))
        self.to_process(osi)


class SP(OpenSeesObject):
    op_base_type = "sp"
    op_type = None

    def __init__(self, osi, node, dof, dof_values):
        self.node = node
        self.dof = int(dof)
        self.dof_values = dof_values

        self._parameters = [self.node.tag, self.dof, *self.dof_values]
        self.to_process(osi)


def analyze(osi, num_inc=1, dt=None, dt_min=None, dt_max=None, jd=None):
    op_type = 'analyze'
    if dt is None:
        parameters = [int(num_inc)]
    elif dt_min is None:
        parameters = [int(num_inc), float(dt)]
    else:
        parameters = [int(num_inc), float(dt), dt_min, dt_max, jd]
    # opy.analyze(*parameters)
    return osi.to_process(op_type, parameters)


def get_node_disp(osi, node, dof):
    op_type = 'nodeDisp'
    parameters = [node.tag, dof]
    # p_str = ', '.join([str(x) for x in parameters])
    return osi.to_process(op_type, parameters)


def get_all_node_disps(osi, dof):
    tags = get_node_tags(osi)
    disps = []
    for tag in tags:
        disps.append(osi.to_process('nodeDisp', [tag, dof]))  # very slow
    return disps


def get_node_vel(osi, node, dof):
    op_type = 'nodeVel'
    parameters = [node.tag, dof]
    return osi.to_process(op_type, parameters)


def get_node_accel(osi, node, dof):
    op_type = 'nodeAccel'
    parameters = [node.tag, dof]
    return osi.to_process(op_type, parameters)


def gen_reactions(osi):
    op_type = 'reactions'
    parameters = []
    return osi.to_process(op_type, parameters)


def get_node_reaction(osi, node, dof):
    op_type = 'nodeReaction'
    parameters = [node.tag, dof]
    return osi.to_process(op_type, parameters)


def get_node_reactions(osi, node):
    op_type = 'nodeReaction'
    parameters = [node.tag]
    return osi.to_process(op_type, parameters)


def get_ele_response(osi, ele, arg, extra_args=None):
    params = [ele.tag, arg]
    if extra_args is not None:
        params += extra_args
    return osi.to_process('eleResponse', params)


def remove_sp(osi, node, dof, pattern=None):
    op_type = 'remove'
    parameters = ['sp', node.tag, dof]
    if pattern is not None:
        parameters.append(pattern.tag)
    # p_str = ', '.join([str(x) for x in parameters])
    return osi.to_process(op_type, parameters)


def remove_load_pattern(osi, load_pattern):
    op_type = 'remove'
    parameters = ['loadPattern', load_pattern.tag]
    return osi.to_process(op_type, parameters)


def remove(osi, o3_obj):
    """Generic remover"""
    op_type = 'remove'
    parameters = [o3_obj.op_base_type, o3_obj.tag]
    return osi.to_process(op_type, parameters)


def set_parameter(osi, value, eles=None, ele_range=None, args=None):
    """Set a parameter on an element"""
    op_type = 'setParameter'
    parameters = ['-val', value]
    if eles is not None:
        ele_tags = [x.tag for x in eles]
        parameters += ['-ele', *ele_tags]
    elif ele_range is not None:
        ele_tags = [x.tag for x in ele_range]
        parameters += ['-eleRange', *ele_tags]
    else:
        raise ValueError("'eles or ele_range must not be None in set_parameter")
    if args:
        parameters += [str(x) for x in args]
    else:
        raise ValueError("'args' can not be None in set_parameter")
    # p_str = ', '.join([str(x) for x in parameters])
    return osi.to_process(op_type, parameters)


def set_time(osi, time):
    """Reset the analysis time to a new value"""
    osi.to_process('setTime', [time])


def get_time(osi):
    """Get the analysis time"""
    return osi.to_process('getTime', [])


def wipe_analysis(osi):
    osi.to_process('wipeAnalysis', [])


def wipe(osi):
    """Wipe the current analysis and save the results to file"""
    osi.to_process('wipe', [])

def reset(osi):
    """Reset the current analysis and save the results to file"""
    osi.to_process('reset', [])


def record(osi):
    """This command is used to cause all the recorders to do a record on the current state of the model."""
    return osi.to_process('record', [])


def load_constant(osi, time=None):
    params = []
    if time is not None:
        params += ['-time', time]
    osi.to_process('loadConst', params)


def update_material_stage(osi, material, stage):
    parameters = ['-material', material.tag, '-stage', stage]
    osi.to_process("updateMaterialStage", parameters)


def get_eigen(osi, solver='genBandArpack', n=1):
    """Gets a list eigen values"""
    parameters = [f'-{solver}', n]
    outs = osi.to_process("eigen", parameters)
    if not hasattr(outs, '__len__'):
        return [outs]
    return outs


def get_pid(osi):
    """Get the processor ID of the calling processor."""
    return osi.to_process('getPID', [])


def get_np(osi):
    """Get total number of processors."""
    return osi.to_process('getNP', [])


def set_num_threads(osi, num):
    """Set the total number of threads"""
    return osi.to_process('setNumThread', [num])


def get_num_threads(osi):
    """return the total number of threads available"""
    return osi.to_process('getNumThread', [])


def get_node_dofs(osi, node):
    """Returns the DOF numbering of a node."""
    return osi.to_process('nodeDOFs', node.tag)


def get_node_tags(osi, mesh=None):
    """Returns the OpenSEES numbering of the nodes."""
    params = []
    if mesh is not None:
        params += ['-mesh', mesh.tag]
    return osi.to_process('getNodeTags', params)


def get_ele_tags(osi, mesh=None):
    """Returns the OpenSEES numbering of the elements."""
    params = []
    if mesh is not None:
        params += ['-mesh', mesh.tag]
    return osi.to_process('getEleTags', params)


def get_node_coords(osi, node, ndm=None):

    if ndm is not None:
        pms = [ndm]
    else:
        pms = []
    return osi.to_process('nodeCoord', [node.tag, *pms])


def get_all_node_coords(osi, ndm=None):
    tags = get_node_tags(osi)
    coords = []
    if ndm is not None:
        pms = [ndm]
    else:
        pms = []
    for tag in tags:
        coords.append(osi.to_process('nodeCoord', [tag, *pms]))  # very slow
    return coords


def get_all_ele_node_tags(osi):
    ele_tags = get_ele_tags(osi)
    node_tags = []
    for tag in ele_tags:
        node_tags.append(osi.to_process('eleNodes', [tag]))
    return node_tags


def get_all_ele_node_tags_as_dict(osi):
    ele_tags = get_ele_tags(osi)
    if not hasattr(ele_tags, '__len__'):
        ele_tags = [ele_tags]
    node_tags = {}
    for ele_tag in ele_tags:
        node_tags[ele_tag] = osi.to_process('eleNodes', [ele_tag])
    return node_tags


def get_all_ele_node_tags_by_n_nodes(osi):
    ele_tags = get_ele_tags(osi)
    all_node_tags = {}
    for ele_tag in ele_tags:
        node_tags = osi.to_process('eleNodes', [ele_tag])
        if node_tags is not None:
            if len(node_tags) not in all_node_tags:
                all_node_tags[len(node_tags)] = {}
            all_node_tags[len(node_tags)][ele_tag] = node_tags
    return all_node_tags


class Parameter(OpenSeesObject):
    op_base_type = 'parameter'
    op_type = None

    def __init__(self, osi, ele, section=None, mat=None, pname=''):
        """

        Parameters
        ----------
        osi
        ele
        section
        mat
        pname
        """
        self.ele = ele
        self.section = section
        self.mat = mat
        self.pname = pname
        osi.n_params += 1
        self._tag = osi.n_params
        self._parameters = [self.tag, 'element', self.ele.tag]
        if self.section is not None:
            self._parameters += ['section', self.section]  # unsure if this is correct
        if self.mat is not None:
            self._parameters += ['material', self.mat]
        self._parameters.append(pname)
        self.to_process(osi)


def update_parameter(osi, param, value):
    """Set a parameter to a new value"""
    op_type = 'updateParameter'
    parameters = [param.tag, value]
    return osi.to_process(op_type, parameters)


def start(osi):
    """Start the timer"""
    return osi.to_process('start', [])

def stop(osi):
    """Stop the timer and print timing information."""
    return osi.to_process('stop', [])


def update_element_domain(osi):
    """Update elements in the domain"""
    return osi.to_process('updateElementDomain', [])

def domain_change(osi):
    return osi.to_process('domainChange', [])


def remesh(osi, alpha):
    return osi.to_process('remesh', [alpha])


from o3seespy.base_model import OpenSeesObject


class TimeSeriesBase(OpenSeesObject):
    op_base_type = "timeSeries"


class Constant(TimeSeriesBase):
    r"""
    The Constant TimeSeries Class
    
    This command is used to construct a TimeSeries object in which the load factor applied remains constant and is
    independent of the time in the domain, i.e. :math:`\lambda = f(t) = C`.
    """
    op_type = 'Constant'

    def __init__(self, osi, factor: float=None):
        """
        Initial method for Constant

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        factor: float, optional
            The load factor applied 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.time_series.Constant(osi, factor=1.0)
        """
        self.osi = osi
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        self.to_process(osi)


class Linear(TimeSeriesBase):
    r"""
    The Linear TimeSeries Class
    
    This command is used to construct a TimeSeries object in which the load factor applied is linearly proportional to
    the time in the domain, i.e.:math:`\lambda = f(t) = cFactor * t`. 
    """
    op_type = 'Linear'

    def __init__(self, osi, factor: float=None):
        """
        Initial method for Linear

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        factor: float, optional
            Linear factor. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.time_series.Linear(osi, factor=1.0)
        """
        self.osi = osi
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        self.to_process(osi)


class Trig(TimeSeriesBase):
    r"""
    The Trig TimeSeries Class
    
    This command is used to construct a TimeSeries object in which the load factor is some trigonemtric function of the
    time in the domain

    .. math::

      \lambda = f(t) = 
      \begin{cases}
          cFactor * sin(\frac{2.0\pi(t-tStart)}{period}+\phi), &  tStart<=t<=tEnd\\
          0.0, & otherwise
      \end{cases}
      \phi = shift - \frac{period}{2.0\pi} * \arcsin(\frac{zeroShift}{cFactor})
    """
    op_type = 'Trig'

    def __init__(self, osi, t_start, t_end, period, factor: float=None, shift: float=None, zero_shift: float=None):
        """
        Initial method for Trig

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        t_start: float
            Starting time of non-zero load factor.
        t_end: float
            Ending time of non-zero load factor.
        period: float
            Characteristic period of sine wave.
        factor: float, optional
            Load factor. 
        shift: float, optional
            Phase shift in radians. 
        zero_shift: float, optional
            Zero shift. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.time_series.Trig(osi, t_start=1.0, t_end=1.0, period=1.0, factor=1.0, shift=0.0, zero_shift=0.0)
        """
        self.osi = osi
        self.t_start = float(t_start)
        self.t_end = float(t_end)
        self.period = float(period)
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        if shift is None:
            self.shift = None
        else:
            self.shift = float(shift)
        if zero_shift is None:
            self.zero_shift = None
        else:
            self.zero_shift = float(zero_shift)
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag, self.t_start, self.t_end, self.period]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        if getattr(self, 'shift') is not None:
            self._parameters += ['-shift', self.shift]
        if getattr(self, 'zero_shift') is not None:
            self._parameters += ['-zeroShift', self.zero_shift]
        self.to_process(osi)


class Triangle(TimeSeriesBase):
    r"""
    The Triangle TimeSeries Class
    
    This command is used to construct a TimeSeries object in which the load factor is some triangular function of the
    time in the domain.

    .. math::

      \lambda = f(t) = 
      \begin{cases}
          slope*k*period+zeroShift, & k < 0.25\\
      cFactor-slope*(k-0.25)*period+zeroShift, & k < 0.75\\
      -cFactor+slope*(k-0.75)*period+zeroShift, & k < 1.0\\
      0.0, & otherwise
      \end{cases}
    

    .. math::

      slope = \frac{cFactor}{period/4}
      k = \frac{t+\phi-tStart}{period}-floor(\frac{t+\phi-tStart}{period})
      \phi = shift - \frac{zeroShift}{slope}
    """
    op_type = 'Triangle'

    def __init__(self, osi, t_start, t_end, period, factor: float=None, shift: float=None, zero_shift: float=None):
        """
        Initial method for Triangle

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        t_start: float
            Starting time of non-zero load factor.
        t_end: float
            Ending time of non-zero load factor.
        period: float
            Characteristic period of sine wave.
        factor: float, optional
            Load factor. 
        shift: float, optional
            Phase shift in radians. 
        zero_shift: float, optional
            Zero shift. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.time_series.Triangle(osi, t_start=1.0, t_end=1.0, period=1.0, factor=1.0, shift=0.0, zero_shift=0.0)
        """
        self.osi = osi
        self.t_start = float(t_start)
        self.t_end = float(t_end)
        self.period = float(period)
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        if shift is None:
            self.shift = None
        else:
            self.shift = float(shift)
        if zero_shift is None:
            self.zero_shift = None
        else:
            self.zero_shift = float(zero_shift)
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag, self.t_start, self.t_end, self.period]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        if getattr(self, 'shift') is not None:
            self._parameters += ['-shift', self.shift]
        if getattr(self, 'zero_shift') is not None:
            self._parameters += ['-zeroShift', self.zero_shift]
        self.to_process(osi)


class Rectangular(TimeSeriesBase):
    r"""
    The Rectangular TimeSeries Class
    
    This command is used to construct a TimeSeries object in which the load factor is constant for a specified period
    and 0 otherwise, i.e.

    .. math::

      \lambda = f(t) = 
      \begin{cases}
          cFactor, &  tStart<=t<=tEnd\\
      0.0, & otherwise
      \end{cases}
    """
    op_type = 'Rectangular'

    def __init__(self, osi, t_start, t_end, factor: float=None):
        """
        Initial method for Rectangular

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        t_start: float
            Starting time of non-zero load factor.
        t_end: float
            Ending time of non-zero load factor.
        factor: float, optional
            Load factor. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.time_series.Rectangular(osi, t_start=1.0, t_end=1.0, factor=1.0)
        """
        self.osi = osi
        self.t_start = float(t_start)
        self.t_end = float(t_end)
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag, self.t_start, self.t_end]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        self.to_process(osi)


class Pulse(TimeSeriesBase):
    r"""
    The Pulse TimeSeries Class
    
    This command is used to construct a TimeSeries object in which the load factor is some pulse function of the time in
    the domain.

    .. math::

      \lambda = f(t) = 
      \begin{cases}
          cFactor+zeroShift, &  k < width\\
      zeroshift, & k < 1\\
      0.0, & otherwise
      \end{cases}
    

    .. math::

      k = \frac{t+shift-tStart}{period}-floor(\frac{t+shift-tStart}{period})
    """
    op_type = 'Pulse'

    def __init__(self, osi, t_start, t_end, period, width: float=None, shift: float=None, factor: float=None, zero_shift: float=None):
        """
        Initial method for Pulse

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        t_start: float
            Starting time of non-zero load factor.
        t_end: float
            Ending time of non-zero load factor.
        period: float
            Characteristic period of pulse.
        width: float, optional
            Pulse width as a fraction of the period. (optinal)
        shift: float, optional
            Phase shift in seconds. 
        factor: float, optional
            Load factor. 
        zero_shift: float, optional
            Zero shift. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.time_series.Pulse(osi, t_start=1.0, t_end=1.0, period=1.0, width=0.5, shift=0.0, factor=1.0, zero_shift=0.0)
        """
        self.osi = osi
        self.t_start = float(t_start)
        self.t_end = float(t_end)
        self.period = float(period)
        if width is None:
            self.width = None
        else:
            self.width = float(width)
        if shift is None:
            self.shift = None
        else:
            self.shift = float(shift)
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        if zero_shift is None:
            self.zero_shift = None
        else:
            self.zero_shift = float(zero_shift)
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag, self.t_start, self.t_end, self.period]
        if getattr(self, 'width') is not None:
            self._parameters += ['-width', self.width]
        if getattr(self, 'shift') is not None:
            self._parameters += ['-shift', self.shift]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        if getattr(self, 'zero_shift') is not None:
            self._parameters += ['-zeroShift', self.zero_shift]
        self.to_process(osi)


class Path(TimeSeriesBase):
    """
    The Path TimeSeries Class
    
    The relationship between loadfactor and time is input by the user as a series of discrete points inthe 2d space
    (load factor, time). The input points can come from afile or from a list in the script. When the time specified
    does not matchany of the input points, linear interpolation is used between points.There are many ways to
    specify the load path, for example,
    """
    op_type = 'Path'

    def __init__(self, osi, dt: float=None, values: list=None, time: list=None, filepath: str=None, file_time: str=None, factor: float=None, start_time: float=None, use_last=False, prepend_zero=False):
        """
        Initial method for Path

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        dt: float, optional
            Time interval between specified points. 
        values: list, optional
            Load factor values in a |list|. 
        time: list, optional
            Time values in a |list|. 
        filepath: str, optional
            File containing the load factors values. 
        file_time: str, optional
            File containing the time values for corresponding load factors. 
        factor: float, optional
            A factor to multiply load factors by. 
        start_time: float, optional
            Provide a start time for provided load factors. 
        use_last: bool
            Use last value after the end of the series. 
        prepend_zero: bool
            Prepend a zero value to the series of load factors. 
        """
        self.osi = osi
        if dt is None:
            self.dt = None
        else:
            self.dt = float(dt)
        self.values = values
        self.time = time
        self.filepath = filepath
        self.file_time = file_time
        if factor is None:
            self.factor = None
        else:
            self.factor = float(factor)
        if start_time is None:
            self.start_time = None
        else:
            self.start_time = float(start_time)
        self.use_last = use_last
        self.prepend_zero = prepend_zero
        osi.n_tseries += 1
        self._tag = osi.n_tseries
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'dt') is not None:
            self._parameters += ['-dt', self.dt]
        if getattr(self, 'values') is not None:
            self._parameters += ['-values', *self.values]
        if getattr(self, 'time') is not None:
            self._parameters += ['-time', *self.time]
        if getattr(self, 'filepath') is not None:
            self._parameters += ['-filepath', self.filepath]
        if getattr(self, 'file_time') is not None:
            self._parameters += ['-fileTime', self.file_time]
        if getattr(self, 'factor') is not None:
            self._parameters += ['-factor', self.factor]
        if getattr(self, 'start_time') is not None:
            self._parameters += ['-startTime', self.start_time]
        if getattr(self, 'use_last'):
            self._parameters += ['-useLast']
        if getattr(self, 'prepend_zero'):
            self._parameters += ['-prependZero']
        self.to_process(osi)



import o3seespy as o3


def test_pressure_depend_multi_yield():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.PressureDependMultiYield(osi, nd=2, rho=1.7, g_mod_ref=6.0e4, bulk_mod_ref=16.0e4, phi=31., peak_strain=0.1, p_ref=101,
                 d=0.5, pt_ang=31., con_rate=0.21, dil_rates=[0., 0.], liquefac=[10, 0.02, 1], n_surf=20.0,
                 e_init=0.85, cs_params=None, c=0.3)


def test_pressure_depend_multi_yield02():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.PressureDependMultiYield02(osi, nd=2, rho=1.7, g_mod_ref=6.0e4, bulk_mod_ref=16.0e4, phi=31.,
                                            peak_strain=0.1, p_ref=101,
                                            d=0.5, pt_ang=31., con_rates=[0.08, 5, 0.18], dil_rates=[0., 3., 0.],
                                            liquefac=[1, 0.], n_surf=20.0,
                                            e_init=0.85, cs_params=None, c=0.3)


from o3seespy.base_model import OpenSeesObject


class UniaxialMaterialBase(OpenSeesObject):
    op_base_type = "uniaxialMaterial"
    op_type = None

    def set_parameter(self, osi, pstr, value, ele, eles):
        from o3seespy import set_parameter
        if ele is not None:
            set_parameter(osi, value=value, eles=[ele], args=[pstr, 1])
        if eles is not None:
            set_parameter(osi, value=value, eles=eles, args=[pstr, 1])

    def build(self, osi):
        self.osi = osi
        osi.n_mat += 1
        self._tag = osi.n_mat
        self.to_process(osi)
        self.built = 1



from o3seespy.command.nd_material.base_material import NDMaterialBase


class ElasticIsotropic(NDMaterialBase):
    """
    The ElasticIsotropic NDMaterial Class
    
    This command is used to construct an ElasticIsotropic material object.
    """
    op_type = 'ElasticIsotropic'

    def __init__(self, osi, e_mod, nu, rho=0.0):
        """
        Initial method for ElasticIsotropic

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Elastic modulus
        nu: float
            Poisson's ratio
        rho: float, optional
            Mass density 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.nu = float(nu)
        self.rho = float(rho)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.e_mod, self.nu, self.rho]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_e_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'E', value, ele, eles)

    def set_v(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'v', value, ele, eles)

    def set_rho(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'rho', value, ele, eles)


class ElasticOrthotropic(NDMaterialBase):
    """
    The ElasticOrthotropic NDMaterial Class
    
    This command is used to construct an ElasticOrthotropic material object.
    """
    op_type = 'ElasticOrthotropic'

    def __init__(self, osi, ex, ey, ez, nu_xy, nu_yz, nu_zx, gxy, gyz, gzx, rho=0.0):
        """
        Initial method for ElasticOrthotropic

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ex: float
            Elastic modulus in x direction
        ey: float
            Elastic modulus in y direction
        ez: float
            Elastic modulus in z direction
        nu_xy: float
            Poisson's ratios in x and y plane
        nu_yz: float
            Poisson's ratios in y and z plane
        nu_zx: float
            Poisson's ratios in z and x plane
        gxy: float
            Shear modulii in x and y plane
        gyz: float
            Shear modulii in y and z plane
        gzx: float
            Shear modulii in z and x plane
        rho: float, optional
            Mass density 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.ElasticOrthotropic(osi, ex=1.0, ey=1.0, ez=1.0, nu_xy=1.0, nu_yz=1.0, nu_zx=1.0, gxy=1.0, gyz=1.0, gzx=1.0, rho=0.0)
        """
        self.osi = osi
        self.ex = float(ex)
        self.ey = float(ey)
        self.ez = float(ez)
        self.nu_xy = float(nu_xy)
        self.nu_yz = float(nu_yz)
        self.nu_zx = float(nu_zx)
        self.gxy = float(gxy)
        self.gyz = float(gyz)
        self.gzx = float(gzx)
        self.rho = float(rho)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.ex, self.ey, self.ez, self.nu_xy, self.nu_yz, self.nu_zx, self.gxy, self.gyz, self.gzx, self.rho]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_e_mod_x(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Ex', value, ele, eles)

    def set_e_mod_y(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Ey', value, ele, eles)

    def set_e_mod_z(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Ez', value, ele, eles)

    def set_vyx(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'vyx', value, ele, eles)

    def set_vzy(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'vzy', value, ele, eles)

    def set_vxz(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'vxz', value, ele, eles)

    def set_g_mod_yx(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Gyx', value, ele, eles)

    def set_g_mod_zy(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Gzy', value, ele, eles)

    def set_g_mod_xz(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Gxz', value, ele, eles)

    def set_rho(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'rho', value, ele, eles)


class J2Plasticity(NDMaterialBase):
    """
    The J2Plasticity NDMaterial Class
    
    This command is used to construct an multi dimensional material object that has a von Mises (J2) yield criterium and
    isotropic hardening.
    """
    op_type = 'J2Plasticity'

    def __init__(self, osi, k_mod, g_mod, sig0, sig_inf, delta, big_h):
        """
        Initial method for J2Plasticity

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        k_mod: float
            Bulk modulus
        g_mod: float
            Shear modulus
        sig0: float
            Initial yield stress
        sig_inf: float
            Final saturation yield stress
        delta: float
            Exponential hardening parameter
        big_h: float
            Linear hardening parameter

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.J2Plasticity(osi, k_mod=1.0, g_mod=1.0, sig0=1.0, sig_inf=1.0, delta=1.0, big_h=1.0)
        """
        self.osi = osi
        self.k_mod = float(k_mod)
        self.g_mod = float(g_mod)
        self.sig0 = float(sig0)
        self.sig_inf = float(sig_inf)
        self.delta = float(delta)
        self.big_h = float(big_h)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.k_mod, self.g_mod, self.sig0, self.sig_inf, self.delta, self.big_h]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_k(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'K', value, ele, eles)

    def set_mu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'mu', value, ele, eles)

    def set_rho(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'rho', value, ele, eles)


class DruckerPrager(NDMaterialBase):
    """
    The DruckerPrager NDMaterial Class
    
    This command is used to construct an multi dimensional material object that has a Drucker-Prager yield criterium.
    """
    op_type = 'DruckerPrager'

    def __init__(self, osi, k_mod, g_mod, sigma_y, rho, rho_bar, kinf, ko, delta1, delta2, big_h, theta, density, atm_pressure=101e3):
        r"""
        Initial method for DruckerPrager

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        k_mod: float
            Bulk modulus
        g_mod: float
            Shear modulus
        sigma_y: float
            Yield stress
        rho: float
            Frictional strength parameter
        rho_bar: float
            Controls evolution of plastic volume change, :math:`0\le rhobar \le rho`.
        kinf: float
            Nonlinear isotropic strain hardening parameter, :math:`kinf \ge 0`.
        ko: float
            Nonlinear isotropic strain hardening parameter, :math:`ko \ge 0`.
        delta1: float
            Nonlinear isotropic strain hardening parameter, :math:`delta1\ge 0`.
        delta2: float
            Tension softening parameter, :math:`delta2\ge 0`.
        big_h: float
            Linear hardening parameter, :math:`h \ge 0`.
        theta: float
            Controls relative proportions of isotropic and kinematic hardening, :math:`0 \le theta \le 1`.
        density: float
            Mass density of the material
        atm_pressure: float, optional
            Optional atmospheric pressure for update of elastic bulk and shear moduli

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.DruckerPrager(osi, k_mod=1.0, g_mod=1.0, sigma_y=1.0, rho=1.0, rho_bar=1.0, kinf=1.0, ko=1.0, delta1=1.0, delta2=1.0, big_h=1.0, theta=1.0, density=1.0, atm_pressure=101e3)
        """
        self.osi = osi
        self.k_mod = float(k_mod)
        self.g_mod = float(g_mod)
        self.sigma_y = float(sigma_y)
        self.rho = float(rho)
        self.rho_bar = float(rho_bar)
        self.kinf = float(kinf)
        self.ko = float(ko)
        self.delta1 = float(delta1)
        self.delta2 = float(delta2)
        self.big_h = float(big_h)
        self.theta = float(theta)
        self.density = float(density)
        self.atm_pressure = float(atm_pressure)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.k_mod, self.g_mod, self.sigma_y, self.rho, self.rho_bar, self.kinf, self.ko, self.delta1, self.delta2, self.big_h, self.theta, self.density, self.atm_pressure]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_material_state(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'materialState', value, ele, eles)

    def set_frictional_strength(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'frictionalStrength', value, ele, eles)

    def set_nonassociative_term(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'nonassociativeTerm', value, ele, eles)

    def set_cohesive_intercept(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'cohesiveIntercept', value, ele, eles)

    def set_g_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'shearModulus', value, ele, eles)

    def set_bulk_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'bulkModulus', value, ele, eles)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)


class Damage2p(NDMaterialBase):
    """
    The Damage2p NDMaterial Class
    
    This command is used to construct a three-dimensional material object that has a Drucker-Prager plasticity model
    coupled with a two-parameter damage model.
    """
    op_type = 'Damage2p'

    def __init__(self, osi, fcc, fct: float=None, e_mod: float=None, ni: float=None, gt: float=None, gc: float=None, rho_bar: float=None, big_h: float=None, theta: float=None, tangent: float=None):
        r"""
        Initial method for Damage2p

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fcc: float
            Concrete compressive strength, negative real value (positive input is changed in sign automatically)
        fct: float, optional
            Optional concrete tensile strength, positive real value (for concrete like materials is less than fcc),
            :math:`0.1*abs(fcc)` = :math:`4750*sqrt(abs(fcc))\text{ }if\text{ }abs(fcc)<2000` because fcc is assumed in mpa
            (see aci 318)
        e_mod: float, optional
            Optional young modulus, :math:`57000*sqrt(abs(fcc))` if :math:`abs(fcc)>2000` because fcc is assumed in psi
            (see aci 318)
        ni: float, optional
            Optional poisson coefficient, 0.15 (from comparison with tests by kupfer hilsdorf rusch 1969)
        gt: float, optional
            Optional tension fracture energy density, positive real value (integral of the stress-strain envelope in
            tension), :math:`1840*fct*fct/e` (from comparison with tests by gopalaratnam and shah 1985)
        gc: float, optional
            Optional compression fracture energy density, positive real value (integral of the stress-strain envelope
            after the peak in compression), :math:6250*fcc*fcc/e` (from comparison with tests by karsan and jirsa 1969)
        rho_bar: float, optional
            Optional parameter of plastic volume change, positive real value :math:`0=rhobar< sqrt(2/3)`, 0.2 (from
            comparison with tests by kupfer hilsdorf rusch 1969)
        big_h: float, optional
            Optional linear hardening parameter for plasticity, positive real value (usually less than e),
            :math:`0.25*e` (from comparison with tests by karsan and jirsa 1969 and gopalaratnam and shah 1985)
        theta: float, optional
            Optional ratio between isotropic and kinematic hardening, positive real value :math:`0=theta=1` (with: 0
            hardening kinematic only and 1 hardening isotropic only, 0.5 (from comparison with tests by karsan and jirsa 1969
            and gopalaratnam and shah 1985)
        tangent: float, optional
            Optional integer to choose the computational stiffness matrix, 0: computational tangent; 1: damaged secant
            stiffness (hint: in case of strong nonlinearities use it with krylov-newton algorithm)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.Damage2p(osi, fcc=1.0, fct=1.0, e_mod=1.0, ni=1.0, gt=1.0, gc=1.0, rho_bar=1.0, big_h=1.0, theta=1.0, tangent=1.0)
        """
        self.osi = osi
        self.fcc = float(fcc)
        if fct is None:
            self.fct = None
        else:
            self.fct = float(fct)
        if e_mod is None:
            self.e_mod = None
        else:
            self.e_mod = float(e_mod)
        if ni is None:
            self.ni = None
        else:
            self.ni = float(ni)
        if gt is None:
            self.gt = None
        else:
            self.gt = float(gt)
        if gc is None:
            self.gc = None
        else:
            self.gc = float(gc)
        if rho_bar is None:
            self.rho_bar = None
        else:
            self.rho_bar = float(rho_bar)
        if big_h is None:
            self.big_h = None
        else:
            self.big_h = float(big_h)
        if theta is None:
            self.theta = None
        else:
            self.theta = float(theta)
        if tangent is None:
            self.tangent = None
        else:
            self.tangent = float(tangent)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fcc]
        if getattr(self, 'fct') is not None:
            self._parameters += ['-fct', self.fct]
        if getattr(self, 'e_mod') is not None:
            self._parameters += ['-E', self.e_mod]
        if getattr(self, 'ni') is not None:
            self._parameters += ['-ni', self.ni]
        if getattr(self, 'gt') is not None:
            self._parameters += ['-Gt', self.gt]
        if getattr(self, 'gc') is not None:
            self._parameters += ['-Gc', self.gc]
        if getattr(self, 'rho_bar') is not None:
            self._parameters += ['-rho_bar', self.rho_bar]
        if getattr(self, 'big_h') is not None:
            self._parameters += ['-H', self.big_h]
        if getattr(self, 'theta') is not None:
            self._parameters += ['-theta', self.theta]
        if getattr(self, 'tangent') is not None:
            self._parameters += ['-tangent', self.tangent]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class PlaneStress(NDMaterialBase):
    """
    The PlaneStress NDMaterial Class
    
    This command is used to construct a plane-stress material wrapper which converts any three-dimensional material into
    a plane stress material via static condensation.
    """
    op_type = 'PlaneStress'

    def __init__(self, osi, mat3d):
        """
        Initial method for PlaneStress

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat3d: obj
            Object of perviously defined 3d ndmaterial material

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> mat_3d = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        >>> o3.nd_material.PlaneStress(osi, mat3d=mat_3d)
        """
        self.osi = osi
        self.mat3d = mat3d
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.mat3d.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_tangent(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Tangent', value, ele, eles)

    def set_stress(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'tangent', value, ele, eles)

    def set_stresses(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stress', value, ele, eles)

    def set_strain(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stresses', value, ele, eles)

    def set_strains(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'strain', value, ele, eles)


class PlaneStrain(NDMaterialBase):
    """
    The PlaneStrain NDMaterial Class
    
    This command is used to construct a plane-stress material wrapper which converts any three-dimensional material into
    a plane strain material by imposing plain strain conditions on the three-dimensional material.
    """
    op_type = 'PlaneStrain'

    def __init__(self, osi, mat3d):
        """
        Initial method for PlaneStrain

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat3d: obj
            Integer object of previously defined 3d ndmaterial material

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat_3d = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        >>> o3.nd_material.PlaneStrain(osi, mat3d=mat_3d)
        """
        self.osi = osi
        self.mat3d = mat3d
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.mat3d.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_tangent(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Tangent', value, ele, eles)

    def set_stress(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'tangent', value, ele, eles)

    def set_stresses(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stress', value, ele, eles)

    def set_strain(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stresses', value, ele, eles)

    def set_strains(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'strain', value, ele, eles)


class MultiaxialCyclicPlasticity(NDMaterialBase):
    """
    The MultiaxialCyclicPlasticity NDMaterial Class
    
    This command is used to construct an multiaxial Cyclic Plasticity model for clays
    """
    op_type = 'MultiaxialCyclicPlasticity'

    def __init__(self, osi, rho, k_mod, g_mod, su, ho, h, m, beta, k_coeff):
        r"""
        Initial method for MultiaxialCyclicPlasticity

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        rho: float
            Density
        k_mod: float
            Buck modulus
        g_mod: float
            Maximum (small strain) shear modulus
        su: float
            Undrained shear strength, size of bounding surface :math:`r=\sqrt{8/3}*su`
        ho: float
            Linear kinematic hardening modulus of bounding surface
        h: float
            Hardening parameter
        m: float
            Hardening parameter
        beta: float
            Integration parameter, usually beta=0.5
        k_coeff: float
            Coefficient of earth pressure, k0

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.MultiaxialCyclicPlasticity(osi, rho=1.0, k_mod=1.0, g_mod=1.0, su=1.0, ho=1.0, h=1.0, m=1.0, beta=1.0, k_coeff=1.0)
        """
        self.osi = osi
        self.rho = float(rho)
        self.k_mod = float(k_mod)
        self.g_mod = float(g_mod)
        self.su = float(su)
        self.ho = float(ho)
        self.h = float(h)
        self.m = float(m)
        self.beta = float(beta)
        self.k_coeff = float(k_coeff)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.rho, self.k_mod, self.g_mod, self.su, self.ho, self.h, self.m, self.beta, self.k_coeff]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class BoundingCamClay(NDMaterialBase):
    """
    The BoundingCamClay NDMaterial Class
    
    This command is used to construct a multi-dimensional bounding surface Cam Clay material object after Borja et al.
    (2001).
    """
    op_type = 'BoundingCamClay'

    def __init__(self, osi, mass_density, big_c, bulk_mod, ocr, mu_o, alpha, lamb, h, m):
        """
        Initial method for BoundingCamClay

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mass_density: float
            Mass density
        big_c: float
            Ellipsoidal axis ratio (defines shape of ellipsoidal loading/bounding surfaces)
        bulk_mod: float
            Initial bulk modulus
        ocr: float
            Overconsolidation ratio
        mu_o: float
            Initial shear modulus
        alpha: float
            Pressure-dependency parameter for modulii (greater than or equal to zero)
        lamb: float
            Soil compressibility index for virgin loading
        h: float
            Hardening parameter for plastic response inside of bounding surface (if h = 0, no hardening)
        m: float
            Hardening parameter (exponent) for plastic response inside of bounding surface (if m = 0, only linear
            hardening)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.BoundingCamClay(osi, mass_density=1.0, big_c=1.0, bulk_mod=1.0, ocr=1.0, mu_o=1.0, alpha=1.0, lamb=1.0, h=1.0, m=1.0)
        """
        self.osi = osi
        self.mass_density = float(mass_density)
        self.big_c = float(big_c)
        self.bulk_mod = float(bulk_mod)
        self.ocr = float(ocr)
        self.mu_o = float(mu_o)
        self.alpha = float(alpha)
        self.lamb = float(lamb)
        self.h = float(h)
        self.m = float(m)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.mass_density, self.big_c, self.bulk_mod, self.ocr, self.mu_o, self.alpha, self.lamb, self.h, self.m]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_material_state(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'materialState', value, ele, eles)


class PlateFiber(NDMaterialBase):
    """
    The PlateFiber NDMaterial Class
    
    This command is used to construct a plate-fiber material wrapper which converts any three-dimensional material into
    a plate fiber material (by static condensation) appropriate for shell analysis.
    """
    op_type = 'PlateFiber'

    def __init__(self, osi, three_d):
        """
        Initial method for PlateFiber

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        three_d: obj
            Material object for a previously-defined three-dimensional material

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        >>> o3.nd_material.PlateFiber(osi, three_d=mat)
        """
        self.osi = osi
        self.three_d = three_d
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.three_d.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_tangent(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Tangent', value, ele, eles)

    def set_stress(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'tangent', value, ele, eles)

    def set_stresses(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stress', value, ele, eles)

    def set_strain(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stresses', value, ele, eles)

    def set_strains(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'strain', value, ele, eles)


class FSAM(NDMaterialBase):
    """
    The FSAM NDMaterial Class
    
    This command is used to construct a nDMaterial FSAM (Fixed-Strut-Angle-Model, Figure 1, Kolozvari et al., 2015),
    which is a plane-stress constitutive model for simulating the behavior of RC panel elements under generalized,
    in-plane, reversed-cyclic loading conditions (Ulugtekin, 2010; Orakcal et al., 2012). In the FSAM
    constitutive model, the strain fields acting on concrete and reinforcing steel components of a
    RC panel are assumed to be equal to each other, implying perfect bond assumption between
    concrete and reinforcing steel bars. While the reinforcing steel bars develop uniaxial
    stresses under strains in their longitudinal direction, the behavior of concrete is
    defined using stress-strain relationships in biaxial directions, the orientation
    of which is governed by the state of cracking in concrete. Although the
    concrete stress-strain relationship used in the FSAM is fundamentally
    uniaxial in nature, it also incorporates biaxial softening effects
    including compression softening and biaxial damage. For transfer
    of shear stresses across the cracks, a friction-based
    elasto-plastic shear aggregate interlock model is
    adopted, together with a linear elastic model
    for representing dowel action on the
    reinforcing steel bars (Kolozvari,
    2013). Note that FSAM
    constitutive model
    is implemented to
    be used with Shear-Flexure Interaction model for RC walls (SFI_MVLEM), but it could be also used elsewhere.
    """
    op_type = 'FSAM'

    def __init__(self, osi, rho, s_x, s_y, conc, rou_x, rou_y, nu, alfadow):
        r"""
        Initial method for FSAM

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        rho: float
            Material density
        s_x: obj
            Object of uniaxial_material simulating horizontal (x) reinforcement
        s_y: obj
            Object of uniaxial_material simulating vertical (y) reinforcement
        conc: obj
            Object of uniaxial_material simulating concrete, shall be used with uniaxial_material concretecm
        rou_x: float
            Reinforcing ratio in horizontal (x) direction (:math:`roux = _{s,x}/a_{gross,x}`)
        rou_y: float
            Reinforcing ratio in vertical (x) direction (:math:`rouy = _{s,y}/a_{gross,y}`)
        nu: float
            Concrete friction coefficient (:math:`0.0 < \nu < 1.5`)
        alfadow: float
            Stiffness coefficient of reinforcement dowel action (:math:`0.0 < alfadow < 0.05`)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=3)
        >>> s_x = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> s_y = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> conc = o3.uniaxial_material.ConcreteCM(osi, fpcc=1.0, epcc=1.0, ec=1.0, rc=1.0, xcrn=1.0, ft=1.0, et=1.0, rt=1.0, xcrp=1.0,
        >>>                                 gap_close=0)
        >>> o3.nd_material.FSAM(osi, rho=1.0, s_x=s_x, s_y=s_y, conc=conc, rou_x=1.0, rou_y=1.0, nu=1.0, alfadow=1.0)
        """
        self.osi = osi
        self.rho = float(rho)
        self.s_x = s_x
        self.s_y = s_y
        self.conc = conc
        self.rou_x = float(rou_x)
        self.rou_y = float(rou_y)
        self.nu = float(nu)
        self.alfadow = float(alfadow)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.rho, self.s_x.tag, self.s_y.tag, self.conc.tag, self.rou_x, self.rou_y, self.nu, self.alfadow]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class ManzariDafalias(NDMaterialBase):
    """
    The ManzariDafalias NDMaterial Class
    
    This command is used to construct a multi-dimensional Manzari-Dafalias(2004) material.
    """
    op_type = 'ManzariDafalias'

    def __init__(self, osi, g0, nu, e_init, mc, c, lambda_c, e0, ksi, p_atm, m, h0, ch, nb, a0, nd, z_max, cz, den):
        r"""
        Initial method for ManzariDafalias

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        g0: float
            Shear modulus constant
        nu: float
            Poisson ratio
        e_init: float
            Initial void ratio
        mc: float
            Critical state stress ratio
        c: float
            Ratio of critical state stress ratio in extension and compression
        lambda_c: float
            Critical state line constant
        e0: float
            Critical void ratio at p = 0
        ksi: float
            Critical state line constant
        p_atm: float
            Atmospheric pressure
        m: float
            Yield surface constant (radius of yield surface in stress ratio space)
        h0: float
            Constant parameter
        ch: float
            Constant parameter
        nb: float
            Bounding surface parameter, :math:`nb \ge 0`
        a0: float
            Dilatancy parameter
        nd: float
            Dilatancy surface parameter :math:`nd \ge 0`
        z_max: float
            Fabric-dilatancy tensor parameter
        cz: float
            Fabric-dilatancy tensor parameter
        den: float
            Mass density of the material

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.ManzariDafalias(osi, g0=1.0, nu=1.0, e_init=1.0, mc=1.0, c=1.0, lambda_c=1.0, e0=1.0, ksi=1.0, p_atm=1.0, m=1.0, h0=1.0, ch=1.0, nb=1.0, a0=1.0, nd=1.0, z_max=1.0, cz=1.0, den=1.0)
        """
        self.osi = osi
        self.g0 = float(g0)
        self.nu = float(nu)
        self.e_init = float(e_init)
        self.mc = float(mc)
        self.c = float(c)
        self.lambda_c = float(lambda_c)
        self.e0 = float(e0)
        self.ksi = float(ksi)
        self.p_atm = float(p_atm)
        self.m = float(m)
        self.h0 = float(h0)
        self.ch = float(ch)
        self.nb = float(nb)
        self.a0 = float(a0)
        self.nd = float(nd)
        self.z_max = float(z_max)
        self.cz = float(cz)
        self.den = float(den)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.g0, self.nu, self.e_init, self.mc, self.c, self.lambda_c, self.e0, self.ksi, self.p_atm, self.m, self.h0, self.ch, self.nb, self.a0, self.nd, self.z_max, self.cz, self.den]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)

    def set_material_state(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'materialState', value, ele, eles)

    def set_integration_scheme(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'IntegrationScheme', value, ele, eles)

    def set_jacobian(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Jacobian', value, ele, eles)

    def set_ref_shear_modulus(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'refShearModulus', value, ele, eles)

    def set_g_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'ShearModulus', value, ele, eles)

    def set_nu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'poissonRatio', value, ele, eles)

    def set_void_ratio(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'voidRatio', value, ele, eles)

    def set_stress_correction(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'stressCorrection', value, ele, eles)


class StressDensity(NDMaterialBase):
    op_type = "stressDensity"

    def __init__(self, osi, den, e_init, big_a, n, nu, a1, b1, a2, b2, a3, b3, fd, mu_0, mu_cyc, sc, big_m, p_atm,
                 ssls=None, hsl=None, ps=None):
        """
        Initial method for StressDensity

        Parameters
        ----------
        m_den: float
            Mass density
        e_not: float
            Initial void ratio
        big_a: float
            Constant for elastic shear modulus
        n: float
            Pressure dependency exponent for elastic shear modulus
        nu: float
            Poisson's ratio
        a1: float
            Peak stress ratio coefficient (:math:`etamax = a1 + b1*is`)
        b1: float
            Peak stress ratio coefficient (:math:`etamax = a1 + b1*is`)
        a2: float
            Max shear modulus coefficient (:math:`gn_max = a2 + b2*is`)
        b2: float
            Max shear modulus coefficient (:math:`gn_max = a2 + b2*is`)
        a3: float
            Min shear modulus coefficient (:math:`gn_min = a3 + b3*is`)
        b3: float
            Min shear modulus coefficient (:math:`gn_min = a3 + b3*is`)
        fd: float
            Degradation constant
        mu_not: float
            Dilatancy coefficient (monotonic loading)
        mu_cyc: float
            Dilatancy coefficient (cyclic loading)
        sc: float
            Dilatancy strain
        big_m: float
            Critical state stress ratio
        patm: float
            Atmospheric pressure (in appropriate units)
        ssls: listf
            Void ratio of quasi steady state (qss-line) at pressures [pmin, 10kpa, 30kpa, 50kpa, 100kpa, 200kpa, 400kpa]
            (default = [0.877, 0.877, 0.873, 0.870, 0.860, 0.850, 0.833])
        hsl: float
            Void ratio of upper reference state (ur-line) for all pressures (default = 0.895)
        p1: float
            Pressure corresponding to ssl1 (default = 1.0 kpa)
        """
        self.osi = osi
        self.den = float(den)
        self.e_init = float(e_init)
        self.big_a = float(big_a)
        self.n = float(n)
        self.nu = float(nu)
        self.a1 = float(a1)
        self.b1 = float(b1)
        self.a2 = float(a2)
        self.b2 = float(b2)
        self.a3 = float(a3)
        self.b3 = float(b3)
        self.fd = float(fd)
        self.mu_0 = float(mu_0)
        self.mu_cyc = float(mu_cyc)
        self.sc = float(sc)
        self.big_m = float(big_m)
        self.p_atm = float(p_atm)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat

        self._parameters = [self.op_type, self._tag, self.den, self.e_init, self.big_a, self.n, self.nu, self.a1, self.b1, self.a2,
                            self.b2, self.a3, self.b3, self.fd, self.mu_0, self.mu_cyc, self.sc, self.big_m,
                            self.p_atm]
        if ssls is not None:
            assert len(ssls) == 7, len(ssls)
            self.ssls = [float(x) for x in ssls]
            if hsl is None:
                self.hsl = 0.895
            else:
                self.hsl = float(hsl)
            self._parameters += [*self.ssls, self.hsl]
        if ps is not None:
            self.ps = [float(x) for x in ps]
            self._parameters += [*self.ps]

        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



class AcousticMedium(NDMaterialBase):
    """
    The AcousticMedium NDMaterial Class
    
    This command is used to construct an acoustic medium NDMaterial object.
    """
    op_type = 'AcousticMedium'

    def __init__(self, osi, k_mod, rho):
        """
        Initial method for AcousticMedium

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        k_mod: float
            Bulk module of the acoustic medium
        rho: float
            Mass density of the acoustic medium

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.AcousticMedium(osi, k_mod=1.0, rho=1.0)
        """
        self.osi = osi
        self.k_mod = float(k_mod)
        self.rho = float(rho)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.k_mod, self.rho]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_kf(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Kf', value, ele, eles)

    def set_rho(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'rho', value, ele, eles)

    def set_gamma(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'gamma', value, ele, eles)



from o3seespy.base_model import OpenSeesObject


class FrictionModelBase(OpenSeesObject):
    op_base_type = "frictionModel"


class Coulomb(FrictionModelBase):
    """
    The Coulomb FrictionModel Class
    
    This command is used to construct a `Coulomb friction <http://en.wikipedia.org/wiki/Friction>`_ model object.
    Coulomb's Law of Friction states that kinetic friction is independent of the sliding velocity.
    """
    op_type = 'Coulomb'

    def __init__(self, osi, mu):
        """
        Initial method for Coulomb

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mu: float
            Coefficient of friction

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.friction_model.Coulomb(osi, mu=1.0)
        """
        self.osi = osi
        self.mu = float(mu)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.mu]
        self.to_process(osi)


class VelDependent(FrictionModelBase):
    """
    The VelDependent FrictionModel Class
    
    This command is used to construct a VelDependent friction model object. It is useful for modeling the behavior of
    `PTFE <http://en.wikipedia.org/wiki/Polytetrafluoroethylene>`_ or PTFE-like materials sliding on a stainless steel
    surface. For a detailed presentation on the velocity dependence of such interfaces please refer to Constantinou
    et al. (1999).
    """
    op_type = 'VelDependent'

    def __init__(self, osi, mu_slow, mu_fast, trans_rate):
        """
        Initial method for VelDependent

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mu_slow: float
            Coefficient of friction at low velocity
        mu_fast: float
            Coefficient of friction at high velocity
        trans_rate: float
            Transition rate from low to high velocity

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.friction_model.VelDependent(osi, mu_slow=1.0, mu_fast=1.0, trans_rate=1.0)
        """
        self.osi = osi
        self.mu_slow = float(mu_slow)
        self.mu_fast = float(mu_fast)
        self.trans_rate = float(trans_rate)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.mu_slow, self.mu_fast, self.trans_rate]
        self.to_process(osi)


class VelNormalFrcDep(FrictionModelBase):
    """
    The VelNormalFrcDep FrictionModel Class
    
    This command is used to construct a VelNormalFrcDep friction model object.
    """
    op_type = 'VelNormalFrcDep'

    def __init__(self, osi, a_slow, n_slow, a_fast, n_fast, alpha0, alpha1, alpha2, max_mu_fact):
        r"""
        Initial method for VelNormalFrcDep

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        a_slow: float
            Constant for coefficient of friction at low velocity
        n_slow: float
            Exponent for coefficient of friction at low velocity
        a_fast: float
            Constant for coefficient of friction at high velocity
        n_fast: float
            Exponent for coefficient of friction at high velocity
        alpha0: float
            Constant rate parameter coefficient
        alpha1: float
            Linear rate parameter coefficient
        alpha2: float
            Quadratic rate parameter coefficient
        max_mu_fact: float
            Factor for determining the maximum coefficient of friction. this value prevents the friction coefficient
            from exceeding an unrealistic maximum value when the normal force becomes very small. the maximum friction
            coefficient is determined from μfast, for example :math:`\mu \leq maxmufac*μfast`.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.friction_model.VelNormalFrcDep(osi, a_slow=1.0, n_slow=1.0, a_fast=1.0, n_fast=1.0, alpha0=1.0, alpha1=1.0, alpha2=1.0, max_mu_fact=1.0)
        """
        self.osi = osi
        self.a_slow = float(a_slow)
        self.n_slow = float(n_slow)
        self.a_fast = float(a_fast)
        self.n_fast = float(n_fast)
        self.alpha0 = float(alpha0)
        self.alpha1 = float(alpha1)
        self.alpha2 = float(alpha2)
        self.max_mu_fact = float(max_mu_fact)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.a_slow, self.n_slow, self.a_fast, self.n_fast, self.alpha0, self.alpha1, self.alpha2, self.max_mu_fact]
        self.to_process(osi)


class VelPressureDep(FrictionModelBase):
    """
    The VelPressureDep FrictionModel Class
    
    This command is used to construct a VelPressureDep friction model object.
    """
    op_type = 'VelPressureDep'

    def __init__(self, osi, mu_slow, mu_fast0, big_a, delta_mu, alpha, trans_rate):
        """
        Initial method for VelPressureDep

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mu_slow: float
            Coefficient of friction at low velocity
        mu_fast0: float
            Initial coefficient of friction at high velocity
        big_a: float
            Nominal contact area
        delta_mu: float
            Pressure parameter calibrated from experimental data
        alpha: float
            Pressure parameter calibrated from experimental data
        trans_rate: float
            Transition rate from low to high velocity

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.friction_model.VelPressureDep(osi, mu_slow=1.0, mu_fast0=1.0, big_a=1.0, delta_mu=1.0, alpha=1.0, trans_rate=1.0)
        """
        self.osi = osi
        self.mu_slow = float(mu_slow)
        self.mu_fast0 = float(mu_fast0)
        self.big_a = float(big_a)
        self.delta_mu = float(delta_mu)
        self.alpha = float(alpha)
        self.trans_rate = float(trans_rate)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.mu_slow, self.mu_fast0, self.big_a, self.delta_mu, self.alpha, self.trans_rate]
        self.to_process(osi)


class VelDepMultiLinear(FrictionModelBase):
    """
    The VelDepMultiLinear FrictionModel Class
    
    This command is used to construct a VelDepMultiLinear friction model object. The friction-velocity relationship is
    given by a multi-linear curve that is define by a set of points. The slope given by the last two specified points on
    the positive velocity axis is extrapolated to infinite positive velocities. Velocity and friction points need to be
    equal or larger than zero (no negative values should be defined). The number of provided velocity points needs to
    be equal to the number of provided friction points.
    """
    op_type = 'VelDepMultiLinear'

    def __init__(self, osi, vel_points: list=None, frn_points: list=None):
        """
        Initial method for VelDepMultiLinear

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        vel_points: list, optional
            List of velocity points along friction-velocity curve
        frn_points: list, optional
            List of friction points along friction-velocity curve

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> vel_points = [0.0, 1.0]
        >>> frn_points = [1.0, 1.0]
        >>> o3.friction_model.VelDepMultiLinear(osi, vel_points=vel_points, frn_points=frn_points)
        """
        self.osi = osi
        self.vel_points = vel_points
        self.frn_points = frn_points
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'vel_points') is not None:
            self._parameters += ['-vel', *self.vel_points]
        if getattr(self, 'frn_points') is not None:
            self._parameters += ['-frn', *self.frn_points]
        self.to_process(osi)



import o3seespy as o3  # for testing only
import pytest


@pytest.mark.skip()
def test_two_node_link():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
    mats = [o3.uniaxial_material.Elastic(osi, 1.0),
            o3.uniaxial_material.Elastic(osi, 1.0)]
    p_delta_vals = [1.0, 1.0]
    o3.element.TwoNodeLink(osi, ele_nodes=ele_nodes, mats=mats, dir=[1, 1], p_delta_vals=p_delta_vals)




from o3seespy import opy


def elastic_section(section_id):

    e_conc = 30e6
    area = 0.3 * 0.4
    inertia = 0.3 * 0.4 ** 3 / 12

    opy.section("Elastic", section_id, *[e_conc, area, inertia])


def bilinear_material(mat_id):
    e_conc = 30e6
    depth = 0.4
    width = 0.3
    inertia = width * depth ** 3 / 12
    ei = e_conc * inertia
    eps_yield = 300.0e6 / 200e9
    phi_y = 2.1 * eps_yield / depth
    mat_props = [ei, 0.05 * ei, phi_y]
    opy.uniaxialMaterial("ElasticBilin", mat_id, *mat_props)


def uniaxial_steel01_material(mat_id):
    mat_type = 'Steel01'

    mat_args = [300e6, 200e9, 0.001]
    opy.uniaxialMaterial(mat_type, mat_id, *mat_args)


def uniaxial_steel01_section(section_id, mat_id=None):
    if mat_id is None:
        mat_id = section_id
    uniaxial_steel01_material(mat_id)

    opy.section("Uniaxial", section_id, mat_id, "Mz")


def define_beam_integration(integ_tag):
    lp_i = 0.1
    lp_j = 0.2
    section1_id = 1
    section2_id = 2
    section3_id = 3
    uniaxial_steel01_section(section1_id)
    uniaxial_steel01_section(section2_id)
    uniaxial_steel01_section(section3_id)
    opy.beamIntegration('HingeMidpoint', integ_tag, section1_id, lp_i, section2_id, lp_j, section3_id)


def define_force_beam_column(ele_i, integ_tag):

    # Establish nodes
    bot_node = 1
    top_node = 2
    transf_tag = 1
    opy.geomTransf('Linear', transf_tag, *[])
    opy.node(bot_node, 0., 0.)
    opy.node(top_node, 0., 5.)
    opy.element('forceBeamColumn', ele_i, bot_node, top_node, transf_tag, integ_tag)



from o3seespy.base_model import OpenSeesObject


class Rayleigh(OpenSeesObject):
    op_base_type = "rayleigh"
    op_type = "rayleigh"

    def __init__(self, osi, alpha_m, beta_k, beta_k_init, beta_k_comm):
        """
        Assign Rayleigh damping to previously defined nodes

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        alpha_m : float
            factor applied to element or nodes mass matrix
        beta_k : float
            factor applied to element current stiffness matrix
        beta_k_init : float
            factor applied to element initial stiffness matrix
        beta_k_comm : float
            factor applied to element committed stiffness matrix
        """
        self.osi = osi
        self.alpha_m = float(alpha_m)
        self.beta_k = float(beta_k)
        self.beta_k_init = float(beta_k_init)
        self.beta_k_comm = float(beta_k_comm)
        self._parameters = [self.alpha_m, self.beta_k, self.beta_k_init, self.beta_k_comm]
        self.to_process(osi)




import numpy as np
import o3seespy as o3


def run_uniaxial_disp_driver(osi, mat_obj, disps, target_d_inc=1.0e-5):
    """
    A Uniaxial material displacement controlled driver

    Parameters
    ----------
    osi: o3.OpenSeesInstance()
        An Opensees instance
    mat_obj: o3.uniaxial_material.UniaxialMaterialBase()
        An instance of uniaxial material
    disps: array_like
        Target displacements
    target_d_inc: float
        Target displacement increment

    Returns
    -------
    disp: array_like
        Actual displacements
    react: array_like
        Reactions at each displacement
    """
    left_node = o3.node.Node(osi, 0, 0)
    right_node = o3.node.Node(osi, 0, 0)
    o3.Fix1DOF(osi, left_node, o3.cc.FIXED)
    o3.Fix1DOF(osi, right_node, o3.cc.FREE)
    ele = o3.element.ZeroLength(osi, [left_node, right_node], mats=[mat_obj], dirs=[o3.cc.DOF2D_X], r_flag=1)

    disp = []
    react = []

    o3.constraints.Plain(osi)
    o3.numberer.RCM(osi)
    o3.system.BandGeneral(osi)
    o3.test_check.NormDispIncr(osi, 0.002, 10, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.integrator.DisplacementControl(osi, right_node, o3.cc.X, -target_d_inc)
    o3.analysis.Static(osi)
    ts_po = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts_po)
    o3.Load(osi, right_node, [1.0])

    disp.append(0)
    react.append(0)
    d_incs = np.diff(disps, prepend=0)
    for i in range(len(disps)):
        d_inc_i = d_incs[i]
        if target_d_inc < abs(d_inc_i):
            n = int(abs(d_inc_i / target_d_inc))
            d_step = d_inc_i / n
        else:
            n = 1
            d_step = d_inc_i
        o3.integrator.DisplacementControl(osi, right_node, o3.cc.X, d_step)
        o3.analyze(osi, n)
        o3.gen_reactions(osi)
        react.append(o3.get_ele_response(osi, ele, 'force')[0])
        end_disp = -o3.get_node_disp(osi, right_node, dof=o3.cc.X)
        disp.append(end_disp)
    return np.array(disp), np.array(react)


def run_uniaxial_force_driver(osi, mat_obj, forces, d_step=0.001, max_steps=10000, handle='silent'):
    """
    A Uniaxial material force-defined driver

    Parameters
    ----------
    osi: o3.OpenSeesInstance()
        An Opensees instance
    mat_obj: o3.uniaxial_material.UniaxialMaterialBase()
        An instance of uniaxial material
    forces: array_like
        Target forces
    d_step: float
        Displacement increment
    max_steps: int
        Maximum number of steps to take to achieve target force
    handle: str
        Behaviour if target force not reached, If 'silent' then  change to next target force,
        if 'warn' then print warning and go to next force,
        else raise error.

    Returns
    -------
    disp: array_like
        Actual displacements
    react: array_like
        Reactions at each displacement
    """
    left_node = o3.node.Node(osi, 0, 0)
    right_node = o3.node.Node(osi, 0, 0)
    o3.Fix1DOF(osi, left_node, o3.cc.FIXED)
    o3.Fix1DOF(osi, right_node, o3.cc.FREE)
    ele = o3.element.ZeroLength(osi, [left_node, right_node], mats=[mat_obj], dirs=[o3.cc.DOF2D_X], r_flag=1)

    o3.constraints.Plain(osi)
    o3.numberer.RCM(osi)
    o3.system.BandGeneral(osi)
    o3.test_check.NormDispIncr(osi, 0.002, 10, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.integrator.DisplacementControl(osi, right_node, o3.cc.X, 0.0001)
    o3.analysis.Static(osi)
    ts_po = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts_po)
    o3.Load(osi, right_node, [1.0])

    react = 0
    disp = [0]
    reacts = [react]

    diffs = np.diff(forces, prepend=0)
    orys = np.where(diffs >= 0, 1, -1)
    for i in range(len(forces)):
        ory = orys[i]
        o3.integrator.DisplacementControl(osi, right_node, o3.cc.X, -d_step * ory)
        for j in range(max_steps):
            if react * ory < forces[i] * ory:
                o3.analyze(osi, 1)
            else:
                break
            o3.gen_reactions(osi)
            react = o3.get_ele_response(osi, ele, 'force')[0]
            reacts.append(react)
            end_disp = -o3.get_node_disp(osi, right_node, dof=o3.cc.X)
            disp.append(end_disp)
        if j == max_steps - 1:
            if handle == 'silent':
                break
            if handle == 'warn':
                print(f'Target force not reached: force={react:.4g}, target: {forces[i]:.4g}')
            else:
                raise ValueError()
    return np.array(disp), np.array(reacts)


def example_run_disp_gen():
    osi = o3.OpenSeesInstance(ndm=1, ndf=1, state=0)
    mat_obj1 = o3.uniaxial_material.PySimple1(osi, 1, 1e3, 0.05, 1.0, 0.0)
    t = np.arange(0, 20, 0.01)
    umax = 0.6
    disps = np.sin(t) * np.arange(len(t)) / len(t) * umax
    disp, react = o3.tools.run_uniaxial_disp_driver(osi, mat_obj1, disps)

    import matplotlib.pyplot as plt
    plt.plot(disp, react)
    plt.show()


def example_run_force_gen():
    osi = o3.OpenSeesInstance(ndm=1, ndf=1, state=0)
    mat_obj1 = o3.uniaxial_material.PySimple1(osi, 1, 1e3, 0.05, 1.0, 0.0)
    forces = [400, 50, 550, -10, 800, -800, 800]
    disp, react = o3.tools.run_uniaxial_force_driver(osi, mat_obj1, forces)

    import matplotlib.pyplot as plt
    plt.plot(disp, react)
    plt.show()



if __name__ == '__main__':
    # example_run_force_gen()
    example_run_disp_gen()


from o3seespy.command.uniaxial_material.base_material import UniaxialMaterialBase



class Elastic(UniaxialMaterialBase):
    """
    The Elastic UniaxialMaterial Class
    
    This command is used to construct an elastic uniaxial material object.
    """
    op_type = 'Elastic'

    def __init__(self, osi, e_mod, eta=0.0, eneg: float=None):
        """
        Initial method for Elastic

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Tangent
        eta: float, optional
            Damping tangent (optional, default=0.0)
        eneg: float (default=True), optional
            Tangent in compression (optional, default=e)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.eta = float(eta)
        if eneg is None:
            self.eneg = None
        else:
            self.eneg = float(eneg)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.e_mod, self.eta]
        special_pms = ['eneg']
        packets = [False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_e_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'E', value, ele, eles)

    def set_epos(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Epos', value, ele, eles)

    def set_eneg(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Eneg', value, ele, eles)

    def set_eta(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'eta', value, ele, eles)


class ElasticPP(UniaxialMaterialBase):
    """
    The ElasticPP UniaxialMaterial Class
    
    This command is used to construct an elastic perfectly-plastic uniaxial material object.
    """
    op_type = 'ElasticPP'

    def __init__(self, osi, e_mod, epsy_p, epsy_n: float=None, eps0=0.0):
        """
        Initial method for ElasticPP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Tangent
        epsy_p: float
            Strain or deformation at which material reaches plastic state in tension
        epsy_n: float (default=True), optional
            Strain or deformation at which material reaches plastic state in compression. (optional, default is tension
            value)
        eps0: float, optional
            Initial strain (optional, default: zero)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.ElasticPP(osi, e_mod=1.0, epsy_p=1.0, epsy_n=None, eps0=0.0)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.epsy_p = float(epsy_p)
        if epsy_n is None:
            self.epsy_n = None
        else:
            self.epsy_n = float(epsy_n)
        self.eps0 = float(eps0)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.e_mod, self.epsy_p]
        special_pms = ['epsy_n', 'eps0']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_fy(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Fy', value, ele, eles)

    def set_e_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'E', value, ele, eles)

    def set_ep(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'ep', value, ele, eles)


class ElasticPPGap(UniaxialMaterialBase):
    """
    The ElasticPPGap UniaxialMaterial Class
    
    This command is used to construct an elastic perfectly-plastic gap uniaxial material object.
    """
    op_type = 'ElasticPPGap'

    def __init__(self, osi, e_mod, fy, gap, eta=0.0, damage='noDamage'):
        """
        Initial method for ElasticPPGap

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Tangent
        fy: float
            Stress or force at which material reaches plastic state
        gap: float
            Initial gap (strain or deformation)
        eta: float, optional
            Hardening ratio (=eh/e), which can be negative
        damage: str, optional
            An optional string to specify whether to accumulate damage or not in the material. with the default
            re-center on load reversal. is provided this recentering will not occur and gap will grow.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.ElasticPPGap(osi, e_mod=1.0, fy=1.0, gap=1.0, eta=0.0, damage='noDamage')
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.fy = float(fy)
        self.gap = float(gap)
        self.eta = float(eta)
        self.damage = damage
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.e_mod, self.fy, self.gap, self.eta, self.damage]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class ENT(UniaxialMaterialBase):
    """
    The ENT UniaxialMaterial Class
    
    This command is used to construct a uniaxial elastic-no tension material object.
    """
    op_type = 'ENT'

    def __init__(self, osi, e_mod):
        """
        Initial method for ENT

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Tangent

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.ENT(osi, e_mod=1.0)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.e_mod]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_e_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'E', value, ele, eles)


class Parallel(UniaxialMaterialBase):
    """
    The Parallel UniaxialMaterial Class
    
    This command is used to construct a parallel material object made up of an arbitrary number of
    previously-constructed UniaxialMaterial objects.
    """
    op_type = 'Parallel'

    def __init__(self, osi, mats, factor_args: list=None):
        """
        Initial method for Parallel

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mats: list
            Identification objects of materials making up the material model
        factor_args: list, optional
            Factors to create a linear combination of the specified materials. factors can be negative to subtract one
            material from an other. (optional, default = 1.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mats = [o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None),
        >>>         o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)]
        >>> factor_args = [1.0, 1.0]
        >>> o3.uniaxial_material.Parallel(osi, mats=mats, factor_args=factor_args)
        """
        self.osi = osi
        self.mats = [x.tag for x in mats]
        self.factor_args = factor_args
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, *self.mats]
        if getattr(self, 'factor_args') is not None:
            self._parameters += ['-factors', *self.factor_args]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Series(UniaxialMaterialBase):
    """
    The Series UniaxialMaterial Class
    
    This command is used to construct a series material object made up of an arbitrary number of previously-constructed
    UniaxialMaterial objects.
    """
    op_type = 'Series'

    def __init__(self, osi, mats):
        """
        Initial method for Series

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mats: list
            Identification objects of materials making up the material model

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mats = [o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None),
        >>>         o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)]
        >>> o3.uniaxial_material.Series(osi, mats=mats)
        """
        self.osi = osi
        self.mats = [x.tag for x in mats]
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, *self.mats]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



import o3seespy as o3  # for testing only


def test_plain():
    osi = o3.OpenSeesInstance(ndm=2)
    ts = o3.time_series.Linear(osi, factor=1.0)
    o3.pattern.Plain(osi, ts=ts, fact=1.0)


def test_uniform_excitation():
    osi = o3.OpenSeesInstance(ndm=2)
    ts = o3.time_series.Linear(osi, factor=1.0)
    o3.pattern.UniformExcitation(osi, dir=1, accel_series=ts, vel0=1.0, fact=1.0)


def test_multiple_support():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.pattern.MultipleSupport(osi)

